"""The configuration module for sphinx."""

import json
import os
import shutil
import sys
from pathlib import Path

# Get metadata
metadata = json.load(os.popen("hatch project metadata"))
project = metadata["name"]
author = ",".join(author["name"] for author in metadata["authors"])
version = metadata["version"]
project_copyright = f"2018-%Y, {author}"

# Copy notebooks
for notebooks in (
    Path("share", "galactic", "core", "closure", "notebooks"),
    Path("share", "galactic", "core", "connection", "notebooks"),
):
    src = Path(sys.prefix, notebooks)
    shutil.copytree(src, notebooks, dirs_exist_ok=True)

# -- General configuration ------------------------------------------------

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.imgconverter",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.mathjax",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx_book_theme",
    "nbsphinx",
    "sphinxcontrib.plantuml",
    "sphinx_autodoc_typehints",
    "sphinx_codeautolink",
    "sphinx_copybutton",
]

copybutton_prompt_text = r">>> |\.\.\. |\$ |In \[\d*\]: | {2,5}\.\.\.: | {5,8}: "
copybutton_prompt_is_regexp = True
copybutton_line_continuation_character = "\\"

plantuml_output_format = "svg_img"

add_module_names = False

# nbsphinx settings
nbsphinx_allow_errors = True

# Napoleon settings
napoleon_google_docstring = True
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = True
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = True
napoleon_use_admonition_for_notes = True
napoleon_use_admonition_for_references = True
napoleon_use_ivar = False
napoleon_use_param = True
napoleon_use_rtype = True

# intersphinx setting
intersphinx_mapping = {"python": ("https://docs.python.org/3", None)}

autodoc_member_order = "bysource"
autodoc_default_options = {
    "member-order": "bysource",
    "members": True,
    "show-inheritance": True,
}

typehints_defaults = "comma"

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
source_suffix = {".rst": "restructuredtext"}

# The master document.
master_doc = "index"

# The language for content by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ["_build", "**.ipynb_checkpoints"]

# The name of the (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "sphinx_book_theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {
    "navigation_depth": 3,
    "logo": {
        "image_dark": "icon-dark.svg",
        "text": project[len("galactic-") :],
    },
    "use_download_button": False,
    "use_fullscreen_button": False,
    "switcher": {
        "version_match": "latest",
        "json_url": "https://galactic.univ-lr.fr/docs/core/closure/versions.json",
    },
    "article_header_end": [
        "article-header-links.html",
        "article-header-buttons.html",
    ],
}

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
html_sidebars = {
    "**": [
        "navbar-logo.html",
        "version-switcher.html",
        "icon-links.html",
        "search-button-field.html",
        "sbt-sidebar-nav.html",
    ],
}
html_show_sourcelink = False
html_favicon = "thegalactic.ico"
html_logo = "icon.svg"
html_title = ""
